{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n# Example: Biased Kalman Filter\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import argparse\nimport itertools\nimport math\nimport os\n\nimport pyro.distributions as dist\nimport torch\nimport torch.nn as nn\nfrom torch.optim import Adam\n\nimport funsor\nimport funsor.ops as ops\nimport funsor.torch.distributions as f_dist\nfrom funsor.domains import Reals\nfrom funsor.pyro.convert import dist_to_funsor, funsor_to_mvn\nfrom funsor.tensor import Tensor, Variable\n\n# We use a 2D continuous-time NCV dynamics model throughout.\n# See http://webee.technion.ac.il/people/shimkin/Estimation09/ch8_target.pdf\nTIME_STEP = 1.0\nNCV_PROCESS_NOISE = torch.tensor(\n    [\n        [1 / 3, 0.0, 1 / 2, 0.0],\n        [0.0, 1 / 3, 0.0, 1 / 2],\n        [1 / 2, 0.0, 1.0, 0.0],\n        [0.0, 1 / 2, 0.0, 1.0],\n    ]\n)\nNCV_TRANSITION_MATRIX = torch.tensor(\n    [\n        [1.0, 0.0, 0.0, 0.0],\n        [0.0, 1.0, 0.0, 0.0],\n        [1.0, 0.0, 1.0, 0.0],\n        [0.0, 1.0, 0.0, 1.0],\n    ]\n)\n\n\n@torch.no_grad()\ndef generate_data(num_frames, num_sensors):\n    \"\"\"\n    Generate data from a damped NCV dynamics model\n    \"\"\"\n    dt = TIME_STEP\n    bias_scale = 4.0\n    obs_noise = 1.0\n    trans_noise = 0.3\n\n    # define dynamics\n    z = torch.cat([10.0 * torch.randn(2), torch.rand(2)])  # position  # velocity\n    damp = 0.1  # damp the velocities\n    f = torch.tensor(\n        [\n            [1, 0, 0, 0],\n            [0, 1, 0, 0],\n            [dt * math.exp(-damp * dt), 0, math.exp(-damp * dt), 0],\n            [0, dt * math.exp(-damp * dt), 0, math.exp(-damp * dt)],\n        ]\n    )\n    trans_dist = dist.MultivariateNormal(\n        torch.zeros(4),\n        scale_tril=trans_noise * torch.linalg.cholesky(NCV_PROCESS_NOISE),\n    )\n\n    # define biased sensors\n    sensor_bias = bias_scale * torch.randn(2 * num_sensors)\n    h = torch.eye(4, 2).unsqueeze(-1).expand(-1, -1, num_sensors).reshape(4, -1)\n    obs_dist = dist.MultivariateNormal(\n        sensor_bias, scale_tril=obs_noise * torch.eye(2 * num_sensors)\n    )\n\n    states = []\n    observations = []\n    for t in range(num_frames):\n        z = z @ f + trans_dist.sample()\n        states.append(z)\n\n        x = z @ h + obs_dist.sample()\n        observations.append(x)\n\n    states = torch.stack(states)\n    observations = torch.stack(observations)\n    assert observations.shape == (num_frames, num_sensors * 2)\n    return observations, states, sensor_bias\n\n\nclass Model(nn.Module):\n    def __init__(self, num_sensors):\n        super(Model, self).__init__()\n        self.num_sensors = num_sensors\n\n        # learnable params\n        self.log_bias_scale = nn.Parameter(torch.tensor(0.0))\n        self.log_obs_noise = nn.Parameter(torch.tensor(0.0))\n        self.log_trans_noise = nn.Parameter(torch.tensor(0.0))\n\n    def forward(self, observations, add_bias=True):\n        obs_dim = 2 * self.num_sensors\n        bias_scale = self.log_bias_scale.exp()\n        obs_noise = self.log_obs_noise.exp()\n        trans_noise = self.log_trans_noise.exp()\n\n        # bias distribution\n        bias = Variable(\"bias\", Reals[obs_dim])\n        assert not torch.isnan(bias_scale), \"bias scales was nan\"\n        bias_dist = dist_to_funsor(\n            dist.MultivariateNormal(\n                torch.zeros(obs_dim),\n                scale_tril=bias_scale * torch.eye(2 * self.num_sensors),\n            )\n        )(value=bias)\n\n        init_dist = dist.MultivariateNormal(\n            torch.zeros(4), scale_tril=100.0 * torch.eye(4)\n        )\n        self.init = dist_to_funsor(init_dist)(value=\"state\")\n\n        # hidden states\n        prev = Variable(\"prev\", Reals[4])\n        curr = Variable(\"curr\", Reals[4])\n        self.trans_dist = f_dist.MultivariateNormal(\n            loc=prev @ NCV_TRANSITION_MATRIX,\n            scale_tril=trans_noise * torch.linalg.cholesky(NCV_PROCESS_NOISE),\n            value=curr,\n        )\n\n        state = Variable(\"state\", Reals[4])\n        obs = Variable(\"obs\", Reals[obs_dim])\n        observation_matrix = Tensor(\n            torch.eye(4, 2)\n            .unsqueeze(-1)\n            .expand(-1, -1, self.num_sensors)\n            .reshape(4, -1)\n        )\n        assert observation_matrix.output.shape == (\n            4,\n            obs_dim,\n        ), observation_matrix.output.shape\n        obs_loc = state @ observation_matrix\n        if add_bias:\n            obs_loc += bias\n        self.observation_dist = f_dist.MultivariateNormal(\n            loc=obs_loc, scale_tril=obs_noise * torch.eye(obs_dim), value=obs\n        )\n\n        logp = bias_dist\n        curr = \"state_init\"\n        logp += self.init(state=curr)\n        for t, x in enumerate(observations):\n            prev, curr = curr, \"state_{}\".format(t)\n            logp += self.trans_dist(prev=prev, curr=curr)\n            logp += self.observation_dist(state=curr, obs=x)\n            # marginalize out previous state\n            logp = logp.reduce(ops.logaddexp, prev)\n        # marginalize out bias variable\n        logp = logp.reduce(ops.logaddexp, \"bias\")\n\n        # save posterior over the final state\n        assert set(logp.inputs) == {\"state_{}\".format(len(observations) - 1)}\n        posterior = funsor_to_mvn(logp, ndims=0)\n\n        # marginalize out remaining variables\n        logp = logp.reduce(ops.logaddexp)\n        assert isinstance(logp, Tensor) and logp.shape == (), logp.pretty()\n        return logp.data, posterior\n\n\ndef track(args):\n    results = {}  # keyed on (seed, bias, num_frames)\n    for seed in args.seed:\n        torch.manual_seed(seed)\n        observations, states, sensor_bias = generate_data(\n            max(args.num_frames), args.num_sensors\n        )\n        for bias, num_frames in itertools.product(args.bias, args.num_frames):\n            print(\n                \"tracking with seed={}, bias={}, num_frames={}\".format(\n                    seed, bias, num_frames\n                )\n            )\n            model = Model(args.num_sensors)\n            optim = Adam(model.parameters(), lr=args.lr, betas=(0.5, 0.8))\n            losses = []\n            for i in range(args.num_epochs):\n                optim.zero_grad()\n                log_prob, posterior = model(observations[:num_frames], add_bias=bias)\n                loss = -log_prob\n                loss.backward()\n                losses.append(loss.item())\n                if i % 10 == 0:\n                    print(loss.item())\n                optim.step()\n\n            # Collect evaluation metrics.\n            final_state_true = states[num_frames - 1]\n            assert final_state_true.shape == (4,)\n            final_pos_true = final_state_true[:2]\n            final_vel_true = final_state_true[2:]\n\n            final_state_est = posterior.loc\n            assert final_state_est.shape == (4,)\n            final_pos_est = final_state_est[:2]\n            final_vel_est = final_state_est[2:]\n            final_pos_error = float(torch.norm(final_pos_true - final_pos_est))\n            final_vel_error = float(torch.norm(final_vel_true - final_vel_est))\n            print(\"final_pos_error = {}\".format(final_pos_error))\n\n            results[seed, bias, num_frames] = {\n                \"args\": args,\n                \"observations\": observations[:num_frames],\n                \"states\": states[:num_frames],\n                \"sensor_bias\": sensor_bias,\n                \"losses\": losses,\n                \"bias_scale\": float(model.log_bias_scale.exp()),\n                \"obs_noise\": float(model.log_obs_noise.exp()),\n                \"trans_noise\": float(model.log_trans_noise.exp()),\n                \"final_state_estimate\": posterior,\n                \"final_pos_error\": final_pos_error,\n                \"final_vel_error\": final_vel_error,\n            }\n        if args.metrics_filename:\n            print(\"saving output to: {}\".format(args.metrics_filename))\n            torch.save(results, args.metrics_filename)\n    return results\n\n\ndef main(args):\n    funsor.set_backend(\"torch\")\n    if (\n        args.force\n        or not args.metrics_filename\n        or not os.path.exists(args.metrics_filename)\n    ):\n        # Increase compression threshold for numerical stability.\n        with funsor.gaussian.Gaussian.set_compression_threshold(3):\n            results = track(args)\n    else:\n        results = torch.load(args.metrics_filename)\n\n    if args.plot_filename:\n        import matplotlib\n\n        matplotlib.use(\"Agg\")\n        import numpy as np\n        from matplotlib import pyplot\n\n        seeds = set(seed for seed, _, _ in results)\n        X = args.num_frames\n        pyplot.figure(figsize=(5, 1.4), dpi=300)\n\n        pos_error = np.array(\n            [\n                [results[s, 0, f][\"final_pos_error\"] for s in seeds]\n                for f in args.num_frames\n            ]\n        )\n        mse = (pos_error**2).mean(axis=1)\n        std = (pos_error**2).std(axis=1) / len(seeds) ** 0.5\n        pyplot.plot(X, mse**0.5, \"k--\")\n        pyplot.fill_between(\n            X, (mse - std) ** 0.5, (mse + std) ** 0.5, color=\"black\", alpha=0.15, lw=0\n        )\n\n        pos_error = np.array(\n            [\n                [results[s, 1, f][\"final_pos_error\"] for s in seeds]\n                for f in args.num_frames\n            ]\n        )\n        mse = (pos_error**2).mean(axis=1)\n        std = (pos_error**2).std(axis=1) / len(seeds) ** 0.5\n        pyplot.plot(X, mse**0.5, \"r-\")\n        pyplot.fill_between(\n            X, (mse - std) ** 0.5, (mse + std) ** 0.5, color=\"red\", alpha=0.15, lw=0\n        )\n\n        pyplot.ylabel(\"Position RMSE\")\n        pyplot.xlabel(\"Track Length\")\n        pyplot.xticks((5, 10, 15, 20, 25, 30))\n        pyplot.xlim(5, 30)\n        pyplot.tight_layout(0)\n        pyplot.savefig(args.plot_filename)\n\n\ndef int_list(args):\n    result = []\n    for arg in args.split(\",\"):\n        if \"-\" in arg:\n            beg, end = map(int, arg.split(\"-\"))\n            result.extend(range(beg, 1 + end))\n        else:\n            result.append(int(arg))\n    return result\n\n\nif __name__ == \"__main__\":\n    parser = argparse.ArgumentParser(description=\"Biased Kalman filter\")\n    parser.add_argument(\n        \"--seed\",\n        default=\"0\",\n        type=int_list,\n        help=\"random seed, comma delimited for multiple runs\",\n    )\n    parser.add_argument(\n        \"--bias\",\n        default=\"0,1\",\n        type=int_list,\n        help=\"whether to model bias, comma deliminted for multiple runs\",\n    )\n    parser.add_argument(\n        \"-f\",\n        \"--num-frames\",\n        default=\"5,10,15,20,25,30\",\n        type=int_list,\n        help=\"number of sensor frames, comma delimited for multiple runs\",\n    )\n    parser.add_argument(\"--num-sensors\", default=5, type=int)\n    parser.add_argument(\"-n\", \"--num-epochs\", default=50, type=int)\n    parser.add_argument(\"--lr\", default=0.1, type=float)\n    parser.add_argument(\"--metrics-filename\", default=\"\", type=str)\n    parser.add_argument(\"--plot-filename\", default=\"\", type=str)\n    parser.add_argument(\"--force\", action=\"store_true\")\n    args = parser.parse_args()\n    main(args)"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.7.9"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}